# This program downloads the table of watershed recreation areas from
# the nyc.gov web site. It parses out the columns of the table, and
# extracts the links to the individual maps. It then downloads the
# individual maps and extracts multilinestrings from them. The
# multilinestrings are imported into a table named, 'intake' in a
# PostGIS database. A secondary query uses the ST_BuildArea function
# to reconstruct multipolygons, tag them with OpenStreetMap attributes
# corresponding to the table columns, and place them all in a single
# table, 'nyc_bws_rec_area.'

package require clock::rfc2822
package require http
package require tdbc::postgres
package require tdom

# URL of the NYC web site, and database connection to be used

set url http://www.nyc.gov/html/dep/pdf/recreation/open_rec_areas.pdf
tdbc::postgres::connection create db -db gis

# The 'intake' table simply contains the multilinestrings from the PDF

catch {
    db allrows {
	drop table intake
    }
}
catch {
    db allrows {
	create table intake( ogc_fid serial primary key,
			     wkb_geometry geometry(multilinestring, 3857) )
    }
} result; puts $result
set clearIntake [db prepare {delete from intake}]

# The 'inLineStr' table contains single linestrings, extracted from all the
# multilinestrings

catch {
    db allrows {
	create temporary table
	    inLineStr( ogc_fid serial primary key,
	               wkb_geometry geometry(linestring, 3857) )
    }
} result; puts $result
set clearInLineStr [db prepare {delete from inLineStr}]

# Statement to homogenize the linestrings imported from PDF

set homogenizeLinestrings [db prepare {
    insert into inLineStr(wkb_geometry)
    select (ST_Dump(
	ST_LineMerge(ST_CollectionHomogenize(ST_Collect( wkb_geometry ORDER BY ogc_fid )))
    )).geom as x from intake
    
}]

set deleteShortLinestrings [db prepare {
    delete from inLineStr where ST_NPoints(wkb_geometry) < 4
}]

# The 'inPoly' table contains polygons assembled from the linestrings,
# which may or may not be valid.

catch {
    db allrows {
	create temporary table
	    inPoly( ogc_fid serial primary key,
	            wkb_geometry geometry(polygon, 3857) )
    }
} result; puts $result
set clearInPoly [db prepare {delete from inPoly}]

# Statement to convert linestrings to polygons

set makePolygons [db prepare  {
    insert into inPoly(wkb_geometry)
    select ST_MakePolygon(wkb_geometry)
    from inLineStr
}]

# The 'inMPoly' table contains multipolygons created from the
# polygons, after forcing them to be valid.

catch {
    db allrows {
	create temporary table
	    inMPoly( ogc_fid serial primary key,
	            wkb_geometry geometry(multipolygon, 3857) )
    }
} result; puts $result
set clearInMPoly [db prepare {delete from inMPoly}]

# Statement to force polygons to be valid, and extract only the
# nondegenerate geometries that result

set forceValid [db prepare {
    insert into inMPoly(wkb_geometry)
    select ST_Multi(ST_CollectionExtract(ST_MakeValid(wkb_geometry), 3)) as g
    from inPoly
}]

# Statement to unpack the multipolygons back to single polygons.

set unpackMulti [db prepare {
    insert into inLineStr(wkb_geometry)
    select ST_ExteriorRing((ST_Dump(wkb_geometry)).geom)
    from inMPoly
}]

# Statement to rebuild the multipolygons from linestrings, with proper
# handling of holes and simplification of the geometry

set buildArea [db prepare {
    insert into inMPoly(wkb_geometry)
    select 
        ST_Multi(ST_Simplify(
	    ST_Buffer(
		ST_Multi(
		    ST_BuildArea(
                        ST_Collect(wkb_geometry) )), -2.5), 2.5))
    from inLineStr
}]

# Schema of the 'nyc_bws_rec_area' table.

catch {
    db allrows {
	drop table nyc_bws_rec_area
    }
}
db allrows {
    create table nyc_bws_rec_area(
	ogc_fid serial primary key,
	wkb_geometry geometry(multipolygon, 3857),
	"name" text,                                   -- Name of the area
	"NYCDEP:town" text,                            -- Township
	"NYSDEC:wildlife_management_unit" varchar(3),  -- Wildlife Management
	                                               -- Unit
	"NYCDEP:last_modified" char(15),               -- UTC date/time last
	                                               -- modified
	"foot" text,                                   -- 'yes' or 'permit'
	                                               -- or listed activities
	"fishing" varchar(10),                         -- 'yes' or 'no'
	"hunting" varchar(10),                         -- 'yes', 'no', 'archery'
	                                               -- 'permit'
	"trapping" varchar(10),                        -- 'yes' 
	"website:map" text,
	           -- URL from which map was downloaded

	-- CONSTANTS

	"website" text,	              -- Link to NYC's watershed recreation site
	"leisure" char(14),                            -- 'nature_reserve'
	"boundary" char(14),                           -- 'protected_area'
	"protect_class" char(2),                    -- 12
	"protection_object" char(5),                   -- 'water'
	"protection_title" text,                 -- 'Watershed Recreation Unit'
	"related_law" text,                            -- law citation
	"operator" text, -- 'New York City Department of Environmental Protection, Bureau of Water Supply'
	"governance_type" char(18),                    -- 'government_managed'
	"site_ownership" char(9))                      -- 'municipal'
}

# Statement to convert the imported multiline into a tagged multipolygon

set convert_stmt [db prepare {
    insert into nyc_bws_rec_area(wkb_geometry, 
				     "name", "NYCDEP:town",
				     "NYSDEC:wildlife_management_unit",
				     "NYCDEP:last_modified",
				     foot, fishing, hunting, trapping,
				     "website:map", website, leisure, boundary,
				     protect_class, protection_object,
				     protection_title, related_law,
				     operator, governance_type, site_ownership)
    select
        wkb_geometry,
        :name, :town, :wmu, :mtimef,
        :sFoot, :sFishing, :sHunting, :sTrapping,
        :url,
	'http://www.nyc.gov/html/dep/html/recreation/index.shtml',
	'nature_reserve',
	'protected_area',
	'12',
	'water',
	'Watershed Recreation Unit',
	'NYCDEP Rules for the Recreational Use of Water Supply Lands and Waters http://www.nyc.gov/html/dep/pdf/recrules/recrules.pdf',
	'New York City Department of Environmental Protection, Bureau of Water Supply',
	'government_managed',
	'municipal'
    from inMPoly
}]

# retrieve_main_pdf --
#
#	Downloads the main PDF with the table of recreation areas
#
# Parameters:
#	url - URL of the page
#
# Results:
#	Returns the path name of the downloaded file
#
# Side effects:
#	The URL is downloaded into a similarly-named file in the working
#	directory,

proc retrieve_main_pdf {url} {

    set fname [file tail $url]
    flush stdout
    set f [open $fname w]
    set token [http::geturl $url -channel $f]
    close $f
    if {[http::status $token] ne {ok}} {
	puts "[file tail $url]: [http::error $token]"
	file delete $fname
	exit 1
    }
    set mtime [clock::rfc2822 parse_date \
		   [dict get [http::meta $token] Last-Modified]]
    http::cleanup $token
    file mtime $fname $mtime
    return $fname
}

# remote_file_is_newer --
#
#	Test whether a remote file is newer than the local copy
#
# Parameters:
#	url - URL of the remote file
#	fname - Path name of the local copy
#
# Results:
#	Returns 1 if the remote file is newer or there is no local copy,
#	returns 0 if there is a local copy with the same modification time
#
# Side effects:
#	Performs a HEAD operation

proc remote_file_is_newer {url fname} {
    puts "check on $url"
    set needed 1
    set status ok
    if {[file exists $fname]} {
	set token [http::geturl $url -validate 1]
	if {[http::status $token] ne {ok}} {
	    set message [http::error $token]
	    puts "[file tail $url]: [http::error $token]"
	    http::cleanup $token
	    return -code error $message
	}
	if {[dict exists [http::meta $token] Last-Modified]} {
	    set mtime [clock::rfc2822 parse_date \
			   [dict get [http::meta $token] Last-Modified]]
	    if {$mtime <= [file mtime $fname]} {
		set needed 0
	    }
	} else {
	    puts "no modification time?"
	    array set x [http::meta $token]; parray x; unset x
	    set needed 0
	}
	http::cleanup $token
    }
    return $needed
}

# retrieve_remote_file --
#
#	Retrieves a remote file and makes a local copy
#
# Parameters:
#	url - URL to retrieve
#	fname - Local file to put the content in
#
# Results:
#	Returns the modification time of the remote file.

proc retrieve_remote_file {url fname} {

    set f [open $fname wb]
    set token [http::geturl $url -channel $f]
    close $f
    if {[http::status $token] ne {ok}} {
	set message [http::error $token]
	file delete $fname
	http::cleanup $token
	return -code error $message
    }
    if {[catch {
	clock::rfc2822 parse_date [dict get [http::meta $token] Last-Modified]
    } mtime]} {
	puts "Couldn't parse Last-Modified in [file tail $url]: $mtime"
	set mtime [clock seconds]
    }
    http::cleanup $token
    file mtime $fname $mtime
    return $mtime
}

# extract_multilinestring_from_pdf --
#
#	Extracts a multilinestring from the GEOPDF and puts it in the 'intake'
#	table.
#
# Parameters:
#	fname - Name of the PDF file
#	layer - Name of the layer within the PDF file
#
# Results:
#	None.
#
# Side effects:
#	Creates the 'intake' table with a single multipolygon

proc extract_multilinestring_from_pdf {fname layer} {
    catch {
	$::clearIntake allrows
    }
    if {[catch {
	exec ogr2ogr -append -t_srs EPSG:3857 -f PostgreSQL \
	    PG:dbname=gis $fname \
	    -nln intake -nlt MULTILINESTRING \
	    $layer
    } result]} {
	puts $result
    }
}

# convert_database_row --
#
#	Makes the key-value pairs for a new row in the database, given
#	the fields in a table row.
#
# Parameters:
#	name - Name of the unit
#	fields - Fields extracted from a row of the master PDF
#	url - URL where the map was retrieved
#	mtime - Modification time of the map
#
# Results:
#	Returns a dictionary with the column-value pairs for a database
#	insertion
#
#	Also adds the dummy keys 'layer' which identifies the layer in the
#	PDF that's expected to have the boundary of the area, and 'dua'
#	which is 1 for designated use areas and 0 for others.

proc convert_database_row {name fields url mtime} {

    # Name the table columns

    lassign $fields \
	town addr wmu paa hike fish hunt trap dua acres
    
    set row {}

    # Set the external metadata

    set name [string trimright $name { *}]
    append name { Unit}
    dict set row name $name
    dict set row url $url
    dict set row mtimef [clock format $mtime -gmt true \
			     -format %Y%m%d%H%M%SZ]
    
    # Set the data from the table cells

    dict set row town $town
    dict set row wmu $wmu
    if {$paa} {
	dict set row layer Layers_PAA
	set access yes
    } else {
	dict set row layer Layers_Permit_Area
	set access permit
    }
    if {$fish} {
	dict set row sFishing $access
    } else {
	dict set row sFishing no
    }
    if {$hunt eq {BOW}} {
	dict set row sHunting archery
    } elseif {$hunt} {
	dict set row sHunting $access
    } else {
	dict set row sHunting no
    }	
    if {$trap} {
	dict set row sTrapping $access
    } else {
	dict set row sTrapping no
    }
    if {$hike} {
	dict set row sFoot $access
    } else {
	set sep {}
	set v {}
	if {$fish} {
	    append v $sep fishing
	    set sep \;
	}
	if {($hunt eq {BOW}) || $hunt} {
	    append v $sep hunting
	    set sep \;
	}
	if {$trap} {
	    append v $sep trapping
	    set sep \;
	}
	dict set row sFoot $v
    }
    if {$dua} {
	dict set row dua 1
    } else {
	dict set row dua 0
    }

    return $row
}

set fname [retrieve_main_pdf $url]
exec pdftohtml $fname >@stdout
set htmlname [file rootname $fname]s.html

file mkdir pdfs

set f [open $htmlname r]
set data [read $f]
close $f
dom parse -html $data doc

foreach elt [$doc getElementsByTagName a] {
    if {[$elt hasAttribute href]} {
	set ref [$elt getAttribute href]
	if {[file extension $ref] eq {.pdf}} {
	    set unit [$elt asText]
	    set url [$elt getAttribute href]
	    flush stdout
	    set fname [file join pdfs [file tail $ref]]
	    if {![remote_file_is_newer $url $fname]} {
		set mtime [file mtime $fname]
	    } else {
		set mtime [retrieve_remote_file $url $fname]
	    }
	    set fields {}
	    while {[llength $fields] < 10} {
		$elt nextSibling elt
		if {[$elt nodeName] eq "a"} break
		if {[set txt [$elt asText]] ne ""} {
		    lappend fields [string trim $txt]
		}
	    }
	    if {[llength $fields] == 10} {

		# Convert the database row to OSM keys and values
		set dbrow [convert_database_row $unit $fields $url $mtime]

		# I can't convert the maps from Designated Use Areas
		if {[dict get $dbrow dua]} continue;

		puts "parse PDF for $unit"
		# Extract the multilinestrings into the 'intake' table'
		extract_multilinestring_from_pdf $fname [dict get $dbrow layer]
		puts "done parsing PDF"

		# Homogenize the multilinestrings into the 'inLineStr' table
		
		$::clearInLineStr allrows
		$::homogenizeLinestrings allrows

		# Temp: What do the linestrings look like?

		puts "Linestrings for $unit:"
		set k 0
		set str {}
		db foreach row {
		    select ogc_fid,
		           ST_NPoints(wkb_geometry) as npoints,
		           ST_AsText(ST_StartPoint(wkb_geometry)) as startpoint,
		           ST_AsText(ST_EndPoint(wkb_geometry)) as endpoint
		    from inLineStr
		} {
		    dict with row {
			if {$startpoint eq $endpoint} {
			    incr k
			} else {
			    puts $row
			    set str {plus }
			}
		    }
		}
		if {$k > 0} {
		    puts "   $str$k complete rings"
		}

		# Drop any linestrings with fewer than three points
		set rs [$::deleteShortLinestrings execute]
		if {[$rs rowcount] > 0} {
		    puts "   of which [$rs rowcount] were degenerate"
		}
		$rs close

		$::clearInPoly allrows
		set rs [$::makePolygons execute]
		puts "   [$rs rowcount] polygons resulted"
		$rs close

		db foreach row {
		    select count(1) as k from inPoly
		    where not ST_IsValid(wkb_geometry)
		} {
		    dict with row {
			if {$k > 0} {
			    puts "   of which $k were invalid"
			}
		    }
		}

		$::clearInMPoly allrows
		set rs [$::forceValid execute]
		puts "    [$rs rowcount] valid multipolygons"
		$rs close

		$::clearInLineStr allrows
		set rs [$::unpackMulti execute]
		puts "    [$rs rowcount] resulting linestrings"
		$rs close

		# Now at last I have linestrings that represent valid rings.
		# Make a proper multipolygon out of them, shrink it, and
		# simplify it.

		$::clearInMPoly allrows
		set rs [$::buildArea execute]
		puts "    [$rs rowcount] multipolygons after cleanup"
		$rs close

		set rs [$::convert_stmt execute $dbrow]
		puts "    [$rs rowcount] rows added to rec area table"
		$rs close
	    }
	}
    }
}

exit ; # The following didn't help
# Now we attempt to repair the topology of the polygons.

puts "dropping topology geometry column"
catch {
    db allrows {
	SELECT topology.DropTopoGeometryColumn('public',
					       'nyc_bws_rec_area', 
					       'topo_geom')
    }
} result; puts $result

puts "dropping topology"
catch {
    db allrows {
        SELECT topology.DropTopology('nyc_wra_topo')
    }
} result; puts $result

puts "creating topology"
catch {
    db allrows {
	SELECT topology.CreateTopology('nyc_wra_topo', 3857)
    }
} result; puts $result

puts "adding topology column"
catch {
    db allrows {
	SELECT topology.AddTopoGeometryColumn('nyc_wra_topo', 
					      'public', 'nyc_bws_rec_area',
					      'topo_geom', 'MULTIPOLYGON')
    }
} result; puts $result

puts "creating topologic geometry"
catch {
    db allrows {
	UPDATE nyc_bws_rec_area
	SET topo_geom = topology.toTopoGeom(wkb_geometry, 'nyc_wra_topo',
					    1, 2.5)
    }
} result; puts $result

